@@ -81,40 +81,103 @@ display its repr.
 """
 from __future__ import annotations
 
+import asyncio
 import inspect
 import math
 import operator
 
 from collections.abc import Iterable, Iterator
+from functools import partial
 from types import FunctionType, MethodType
 from typing import Any, Callable, Optional
 
-from . import Event
 from .depends import depends
 from .display import _display_accessors, _reactive_display_objs
 from .parameterized import (
-    Parameter, Parameterized, eval_function_with_deps, get_method_owner,
+    Parameter, Parameterized, Skip, Undefined, eval_function_with_deps, get_method_owner,
     register_reference_transform, resolve_ref, resolve_value, transform_reference
 )
-from ._utils import iscoroutinefunction, full_groupby
+from .parameters import Boolean, Event
+from ._utils import _to_async_gen, iscoroutinefunction, full_groupby
 
 
 class Wrapper(Parameterized):
     """
-    Simple wrapper to allow updating literal values easily.
+    Helper class to allow updating literal values easily.
     """
 
     object = Parameter(allow_refs=False)
 
 
+class GenWrapper(Parameterized):
+    """
+    Helper class to allow streaming from generator functions.
+    """
+
+    object = Parameter(allow_refs=True)
+
+
 class Trigger(Parameterized):
+    """
+    Helper class to allow triggering an event under some condition.
+    """
 
     value = Event()
 
-    def __init__(self, parameters, **params):
+    def __init__(self, parameters=None, internal=False, **params):
         super().__init__(**params)
+        self.internal = internal
         self.parameters = parameters
 
+class Resolver(Parameterized):
+    """
+    Helper class to allow (recursively) resolving references.
+    """
+
+    object = Parameter(allow_refs=True)
+
+    recursive = Boolean(default=False)
+
+    value = Parameter()
+
+    def __init__(self, **params):
+        self._watchers = []
+        super().__init__(**params)
+
+    def _resolve_value(self, *events):
+        nested = self.param.object.nested_refs
+        refs = resolve_ref(self.object, nested)
+        value = resolve_value(self.object, nested)
+        if self.recursive:
+            new_refs = [r for r in resolve_ref(value, nested) if r not in refs]
+            while new_refs:
+                refs += new_refs
+                value = resolve_value(value, nested)
+                new_refs = [r for r in resolve_ref(value, nested) if r not in refs]
+            if events:
+                self._update_refs(refs)
+        self.value = value
+        return refs
+
+    @depends('object', watch=True, on_init=True)
+    def _resolve_object(self):
+        refs = self._resolve_value()
+        self._update_refs(refs)
+
+    def _update_refs(self, refs):
+        for w in self._watchers:
+            (w.inst or w.cls).param.unwatch(w)
+        self._watchers = []
+        for _, params in full_groupby(refs, lambda x: id(x.owner)):
+            self._watchers.append(
+                params[0].owner.param.watch(self._resolve_value, [p.name for p in params])
+            )
+
+
+class NestedResolver(Resolver):
+
+    object = Parameter(allow_refs=True, nested_refs=True)
+
 
 class reactive_ops:
     """
@@ -127,46 +190,100 @@ class reactive_ops:
     def __init__(self, reactive):
         self._reactive = reactive
 
+    def _as_rx(self):
+        return self._reactive if isinstance(self._reactive, rx) else self()
+
     def __call__(self):
         rxi = self._reactive
         return rxi if isinstance(rx, rx) else rx(rxi)
 
+    def and_(self, other):
+        """
+        Replacement for the ``and`` statement.
+        """
+        return self._as_rx()._apply_operator(lambda obj, other: obj and other, other)
+
     def bool(self):
         """
         __bool__ cannot be implemented so it is provided as a method.
         """
-        rxi = self._reactive if isinstance(self._reactive, rx) else self()
-        return rxi._apply_operator(bool)
+        return self._as_rx()._apply_operator(bool)
+
+    def buffer(self, n):
+        """
+        Collects the last n items that were emmitted.
+        """
+        items = []
+        def collect(new, n):
+            items.append(new)
+            while len(items) > n:
+                items.pop(0)
+            return items
+        return self._as_rx()._apply_operator(collect, n)
 
     def in_(self, other):
         """
         Replacement for the ``in`` statement.
         """
-        rxi = self._reactive if isinstance(self._reactive, rx) else self()
-        return rxi._apply_operator(operator.contains, other, reverse=True)
+        return self._as_rx()._apply_operator(operator.contains, other, reverse=True)
 
     def is_(self, other):
         """
         Replacement for the ``is`` statement.
         """
-        rxi = self._reactive if isinstance(self._reactive, rx) else self()
-        return rxi._apply_operator(operator.is_, other)
+        return self._as_rx()._apply_operator(operator.is_, other)
 
     def is_not(self, other):
         """
         Replacement for the ``is not`` statement.
         """
-        rxi = self._reactive if isinstance(self._reactive, rx) else self()
-        return rxi._apply_operator(operator.is_not, other)
+        return self._as_rx()._apply_operator(operator.is_not, other)
 
     def len(self):
         """
         __len__ cannot be implemented so it is provided as a method.
         """
-        rxi = self._reactive if isinstance(self._reactive, rx) else self()
-        return rxi._apply_operator(len)
+        return self._as_rx()._apply_operator(len)
 
-    def pipe(self, func, *args, **kwargs):
+    def map(self, func, /, *args, **kwargs):
+        """
+        Apply a function to each item.
+
+        Arguments
+        ---------
+        func: function
+          Function to apply.
+        args: iterable, optional
+          Positional arguments to pass to `func`.
+        kwargs: mapping, optional
+          A dictionary of keywords to pass to `func`.
+        """
+        if inspect.isasyncgenfunction(func) or inspect.isgeneratorfunction(func):
+            raise TypeError(
+                "Cannot map a generator function. Only regular function "
+                "or coroutine functions are permitted."
+            )
+        if inspect.iscoroutinefunction(func):
+            async def apply(vs, *args, **kwargs):
+                return list(await asyncio.gather(*(func(v, *args, **kwargs) for v in vs)))
+        else:
+            def apply(vs, *args, **kwargs):
+                return [func(v, *args, **kwargs) for v in vs]
+        return self._as_rx()._apply_operator(apply, *args, **kwargs)
+
+    def not_(self):
+        """
+        __bool__ cannot be implemented so not has to be provided as a method.
+        """
+        return self._as_rx()._apply_operator(operator.not_)
+
+    def or_(self, other):
+        """
+        Replacement for the ``or`` statement.
+        """
+        return self._as_rx()._apply_operator(lambda obj, other: obj or other, other)
+
+    def pipe(self, func, /, *args, **kwargs):
         """
         Apply chainable functions.
 
@@ -179,8 +296,28 @@ class reactive_ops:
         kwargs: mapping, optional
           A dictionary of keywords to pass to `func`.
         """
-        rxi = self._reactive if isinstance(self._reactive, rx) else self()
-        return rxi._apply_operator(func, *args, **kwargs)
+        return self._as_rx()._apply_operator(func, *args, **kwargs)
+
+    def resolve(self, nested=True, recursive=False):
+        """
+        Resolves references held by the expression.
+
+        As an example if the expression returns a list of parameters
+        this operation will return a list of the parameter values.
+
+        Arguments
+        ---------
+        nested: bool
+          Whether to resolve references contained within nested objects,
+          i.e. tuples, lists, sets and dictionaries.
+        recursive: bool
+          Whether to recursively resolve references, i.e. if a reference
+          itself returns a reference we recurse into it until no more
+          references can be resolved.
+        """
+        resolver_type = NestedResolver if nested else Resolver
+        resolver = resolver_type(object=self._reactive, recursive=recursive)
+        return resolver.param.value.rx()
 
     def updating(self):
         """
@@ -191,17 +328,32 @@ class reactive_ops:
         self._watch(lambda e: wrapper.param.update(object=False), precedence=999)
         return wrapper.param.object.rx()
 
-    def when(self, *dependencies):
+    def when(self, *dependencies, initial=Undefined):
         """
         Returns a reactive expression that emits the contents of this
-        expression only when the condition changes.
+        expression only when the dependencies change. If initial value
+        is provided and the dependencies are all param.Event types the
+        expression will not be evaluated until the first event is
+        triggered.
 
         Arguments
         ---------
         dependencies: param.Parameter | rx
           A dependency that will trigger an update in the output.
+        initial: object
+          Object that will stand in for the actual value until the
+          first time a param.Event in the dependencies is triggered.
         """
-        return bind(lambda *_: self.value, *dependencies).rx()
+        deps = [p for d in dependencies for p in resolve_ref(d)]
+        is_event = all(isinstance(dep, Event) for dep in deps)
+        def eval(*_, evaluated=[]):
+            if is_event and initial is not Undefined and not evaluated:
+                # Abuse mutable default value to keep track of evaluation state
+                evaluated.append(True)
+                return initial
+            else:
+                return self.value
+        return bind(eval, *deps).rx()
 
     def where(self, x, y):
         """
@@ -283,9 +435,11 @@ class reactive_ops:
             )
         self._reactive._wrapper.object = resolve_value(new)
 
-    def watch(self, fn, onlychanged=True, queued=False, precedence=0):
+    def watch(self, fn=None, onlychanged=True, queued=False, precedence=0):
         """
-        Adds a callback that observes the output of the pipeline.
+        Adds a callable that observes the output of the pipeline.
+        If no callable is provided this simply causes the expression
+        to be eagerly evaluated.
         """
         if precedence < 0:
             raise ValueError("User-defined watch callbacks must declare "
@@ -293,19 +447,14 @@ class reactive_ops:
                              "are reserved for internal Watchers.")
         self._watch(fn, onlychanged=onlychanged, queued=queued, precedence=precedence)
 
-    def _watch(self, fn, onlychanged=True, queued=False, precedence=0):
-        def cb(*args):
-            fn(self.value)
-
-        if isinstance(self._reactive, rx):
-            params = self._reactive._params
-        else:
-            params = resolve_ref(self._reactive)
-        for _, group in full_groupby(params, lambda x: id(x.owner)):
-            group[0].owner.param._watch(
-                cb, [dep.name for dep in group], onlychanged=onlychanged,
-                queued=queued, precedence=precedence
-            )
+    def _watch(self, fn=None, onlychanged=True, queued=False, precedence=0):
+        def cb(value):
+            from .parameterized import async_executor
+            if iscoroutinefunction(fn):
+                async_executor(partial(fn, value))
+            elif fn is not None:
+                fn(value)
+        bind(cb, self._reactive, watch=True)
 
 
 def bind(function, *args, watch=False, **kwargs):
@@ -386,9 +535,13 @@ def bind(function, *args, watch=False, **kwargs):
         for kw, arg in wkwargs.items():
             if asynchronous:
                 if kw.startswith('__arg'):
-                    combined_args[int(kw[5:])] = arg
+                    index = kw[5:]
+                    if index.isdigit():
+                        combined_args[int(index)] = arg
                 elif kw.startswith('__kwarg'):
-                    combined_kwargs[kw[8:]] = arg
+                    substring = kw[8:]
+                    if substring in combined_kwargs:
+                        combined_kwargs[substring] = arg
                 continue
             elif kw.startswith('__arg') or kw.startswith('__kwarg') or kw.startswith('__fn'):
                 continue
@@ -406,7 +559,17 @@ def bind(function, *args, watch=False, **kwargs):
                 fn = eval_function_with_deps(p)
         return fn
 
-    if inspect.isasyncgenfunction(function):
+    if inspect.isgeneratorfunction(function):
+        def wrapped(*wargs, **wkwargs):
+            combined_args, combined_kwargs = combine_arguments(
+                wargs, wkwargs, asynchronous=True
+            )
+            evaled = eval_fn()(*combined_args, **combined_kwargs)
+            for val in evaled:
+                yield val
+        wrapper_fn = depends(**dependencies, watch=watch)(wrapped)
+        wrapped._dinfo = wrapper_fn._dinfo
+    elif inspect.isasyncgenfunction(function):
         async def wrapped(*wargs, **wkwargs):
             combined_args, combined_kwargs = combine_arguments(
                 wargs, wkwargs, asynchronous=True
@@ -516,19 +679,28 @@ class rx:
         """
         cls._method_handlers[method] = handler
 
-    def __new__(cls, obj, **kwargs):
+    def __new__(cls, obj=None, **kwargs):
         wrapper = None
         obj = transform_reference(obj)
         if kwargs.get('fn'):
+            # rx._clone codepath
             fn = kwargs.pop('fn')
             wrapper = kwargs.pop('_wrapper', None)
+        elif inspect.isgeneratorfunction(obj) or iscoroutinefunction(obj):
+            # Resolves generator and coroutine functions lazily
+            wrapper = GenWrapper(object=obj)
+            fn = bind(lambda obj: obj, wrapper.param.object)
+            obj = Undefined
         elif isinstance(obj, (FunctionType, MethodType)) and hasattr(obj, '_dinfo'):
+            # Bound functions and methods are resolved on access
             fn = obj
-            obj = eval_function_with_deps(obj)
+            obj = None
         elif isinstance(obj, Parameter):
             fn = bind(lambda obj: obj, obj)
             obj = getattr(obj.owner, obj.name)
         else:
+            # For all other objects wrap them so they can be updated
+            # via .rx.value property
             wrapper = Wrapper(object=obj)
             fn = bind(lambda obj: obj, wrapper.param.object)
         inst = super(rx, cls).__new__(cls)
@@ -538,7 +710,7 @@ class rx:
         return inst
 
     def __init__(
-        self, obj, operation=None, fn=None, depth=0, method=None, prev=None,
+        self, obj=None, operation=None, fn=None, depth=0, method=None, prev=None,
         _shared_obj=None, _current=None, _wrapper=None, **kwargs
     ):
         # _init is used to prevent to __getattribute__ to execute its
@@ -558,19 +730,27 @@ class rx:
         self._depth = depth
         self._dirty = _current is None
         self._dirty_obj = False
+        self._current_task = None
         self._error_state = None
         self._current_ = _current
         if isinstance(obj, rx) and not prev:
             self._prev = obj
         else:
             self._prev = prev
+
+        # Define special trigger parameter if operation has to be lazily evaluated
+        if operation and (iscoroutinefunction(operation['fn']) or inspect.isgeneratorfunction(operation['fn'])):
+            self._trigger = Trigger(internal=True)
+            self._current_ = Undefined
+        else:
+            self._trigger = None
         self._root = self._compute_root()
         self._fn_params = self._compute_fn_params()
         self._internal_params = self._compute_params()
         # Filter params that external objects depend on, ensuring
         # that Trigger parameters do not cause double execution
         self._params = [
-            p for p in self._internal_params if not isinstance(p.owner, Trigger)
+            p for p in self._internal_params if (not isinstance(p.owner, Trigger) or p.owner.internal)
             or any (p not in self._internal_params for p in p.owner.parameters)
         ]
         self._setup_invalidations(depth)
@@ -633,7 +813,9 @@ class rx:
         return args + kwargs
 
     def _compute_params(self) -> list[Parameter]:
-        ps = self._fn_params
+        ps = list(self._fn_params)
+        if self._trigger:
+            ps.append(self._trigger.param.value)
 
         # Collect parameters on previous objects in chain
         prev = self._prev
@@ -647,6 +829,9 @@ class rx:
             return ps
 
         # Accumulate dependencies in args and/or kwargs
+        for ref in resolve_ref(self._operation['fn']):
+            if ref not in ps:
+                ps.append(ref)
         for arg in list(self._operation['args'])+list(self._operation['kwargs'].values()):
             for ref in resolve_ref(arg):
                 if ref not in ps:
@@ -675,11 +860,15 @@ class rx:
         """
         if self._fn is not None:
             for _, params in full_groupby(self._fn_params, lambda x: id(x.owner)):
-                params[0].owner.param._watch(self._invalidate_obj, [p.name for p in params], precedence=-1)
+                fps = [p.name for p in params if p in self._root._fn_params]
+                if fps:
+                    params[0].owner.param._watch(self._invalidate_obj, fps, precedence=-1)
         for _, params in full_groupby(self._internal_params, lambda x: id(x.owner)):
             params[0].owner.param._watch(self._invalidate_current, [p.name for p in params], precedence=-1)
 
     def _invalidate_current(self, *events):
+        if all(event.obj is self._trigger for event in events):
+            return
         self._dirty = True
         self._error_state = None
 
@@ -687,15 +876,46 @@ class rx:
         self._root._dirty_obj = True
         self._error_state = None
 
+    async def _resolve_async(self, obj):
+        self._current_task = task = asyncio.current_task()
+        if inspect.isasyncgen(obj):
+            async for val in obj:
+                if self._current_task is not task:
+                    break
+                self._current_ = val
+                self._trigger.param.trigger('value')
+        else:
+            value = await obj
+            if self._current_task is task:
+                self._current_ = value
+                self._trigger.param.trigger('value')
+
+    def _lazy_resolve(self, obj):
+        from .parameterized import async_executor
+        if inspect.isgenerator(obj):
+            obj = _to_async_gen(obj)
+        async_executor(partial(self._resolve_async, obj))
+
     def _resolve(self):
         if self._error_state:
             raise self._error_state
         elif self._dirty or self._root._dirty_obj:
             try:
                 obj = self._obj if self._prev is None else self._prev._resolve()
+                if obj is Skip or obj is Undefined:
+                    self._current_ = Undefined
+                    raise Skip
                 operation = self._operation
                 if operation:
                     obj = self._eval_operation(obj, operation)
+                    if inspect.isasyncgen(obj) or inspect.iscoroutine(obj) or inspect.isgenerator(obj):
+                        self._lazy_resolve(obj)
+                        obj = Skip
+                    if obj is Skip:
+                        raise Skip
+            except Skip:
+                self._dirty = False
+                return self._current_
             except Exception as e:
                 self._error_state = e
                 raise e
@@ -799,7 +1019,7 @@ class rx:
         # Getting all the public attributes available on the current object,
         # e.g. `sum`, `head`, etc.
         extras = [d for d in dir(current) if not d.startswith('_')]
-        if name in extras and name not in super().__dir__():
+        if (name in extras or current is Undefined) and name not in super().__dir__():
             new = self._resolve_accessor()
             # Setting the method name for a potential use later by e.g. an
             # operator or method, as in `dfi.A > 2`. or `dfi.A.max()`
@@ -875,8 +1095,6 @@ class rx:
         return self._apply_operator(operator.inv)
     def __neg__(self):
         return self._apply_operator(operator.neg)
-    def __not__(self):
-        return self._apply_operator(operator.not_)
     def __pos__(self):
         return self._apply_operator(operator.pos)
     def __trunc__(self):
@@ -979,10 +1197,16 @@ class rx:
         fn, args, kwargs = operation['fn'], operation['args'], operation['kwargs']
         resolved_args = []
         for arg in args:
-            resolved_args.append(resolve_value(arg))
+            val = resolve_value(arg)
+            if val is Skip or val is Undefined:
+                raise Skip
+            resolved_args.append(val)
         resolved_kwargs = {}
         for k, arg in kwargs.items():
-            resolved_kwargs[k] = resolve_value(arg)
+            val = resolve_value(arg)
+            if val is Skip or val is Undefined:
+                raise Skip
+            resolved_kwargs[k] = val
         if isinstance(fn, str):
             obj = getattr(obj, fn)(*resolved_args, **resolved_kwargs)
         elif operation.get('reverse'):
